#include <z3++.h>
#include <iostream>

int main() {
    try {
        // Initialize Z3 context
        z3::context ctx;
        
        z3::sort Element = ctx.uninterpreted_sort("Element");

        z3::sort_vector a_args(ctx);
        a_args.push_back(Element);
        a_args.push_back(Element);
        z3::func_decl a = ctx.function("a", a_args, Element);

        z3::sort_vector b_args(ctx);
        b_args.push_back(Element);
        b_args.push_back(Element);
        z3::func_decl b = ctx.function("b", b_args, Element);

        z3::sort_vector c_args(ctx);
        c_args.push_back(Element);
        c_args.push_back(Element);
        z3::func_decl c = ctx.function("c", c_args, Element);

        z3::sort_vector d_args(ctx);
        d_args.push_back(Element);
        d_args.push_back(Element);
        z3::func_decl d = ctx.function("d", d_args, Element);

        // Create variables X, Y, Z, W as constants of type 'Element'
        z3::expr X = ctx.constant("X", Element);
        z3::expr Y = ctx.constant("Y", Element);
        z3::expr Z = ctx.constant("Z", Element);
        z3::expr W = ctx.constant("W", Element);

        // Initialize solver
        z3::solver solver(ctx);

        // Axiom 1: a(X,X) = X & b(X,X) = X & c(X,X) = X & d(X,X) = X
        z3::expr axiom_idents = (a(X, X) == X) &&
                                 (b(X, X) == X) &&
                                 (c(X, X) == X) &&
                                 (d(X, X) == X);
        solver.add(axiom_idents);

        // Axiom 2: ∀X,Y,Z,W. (a(X,Y) = Z ∧ a(Z,X) = W) ⇒ a(X,W) = Y
        z3::expr axiom_eq1323_00 = z3::implies((a(X, Y) == Z) && (a(Z, X) == W), (a(X, W) == Y));
        solver.add(axiom_eq1323_00);

        // Axiom 3: ∀X,Y,Z,W. (a(X,Y) = Z ∧ a(W,Y) = Z) ⇒ X = W
        z3::expr axiom_a_right_cancellative = z3::implies((a(X, Y) == Z) && (a(W, Y) == Z), (X == W));
        solver.add(axiom_a_right_cancellative);

        // Axiom 4: ∀X,Y,Z,W. (a(X,Y) = Z ∧ a(X,W) = Z) ⇒ Y = W
        z3::expr axiom_a_left_cancellative = z3::implies((a(X, Y) == Z) && (a(X, W) == Z), (Y == W));
        solver.add(axiom_a_left_cancellative);

        // Axiom 5: ∀X,Y,Z,W. (b(X,Y) = Z ∧ c(Z,X) = W) ⇒ b(X,W) = Y
        z3::expr axiom_eq1323_01 = z3::implies((b(X, Y) == Z) && (c(Z, X) == W), (b(X, W) == Y));
        solver.add(axiom_eq1323_01);

        // Axiom 6: ∀X,Y,Z,W. (a(X,Y) = Z ∧ b(Z,X) = W) ⇒ d(X,W) = Y
        z3::expr axiom_eq1323_10 = z3::implies((a(X, Y) == Z) && (b(Z, X) == W), (d(X, W) == Y));
        solver.add(axiom_eq1323_10);

        // Axiom 7: ∀X,Y,Z,W. (b(X,Y) = Z ∧ d(Z,X) = W) ⇒ c(X,W) = Y
        z3::expr axiom_eq1323_11 = z3::implies((b(X, Y) == Z) && (d(Z, X) == W), (c(X, W) == Y));
        solver.add(axiom_eq1323_11);

        // Axiom 8: ∀X,Y,Z,W. (b(X,Y) = Z ∧ b(W,Y) = Z) ⇒ X = W
        z3::expr axiom_b_right_cancellative = z3::implies((b(X, Y) == Z) && (b(W, Y) == Z), (X == W));
        solver.add(axiom_b_right_cancellative);

        // Axiom 9: ∀X,Y,Z,W. (b(X,Y) = Z ∧ b(X,W) = Z) ⇒ Y = W
        z3::expr axiom_b_left_cancellative = z3::implies((b(X, Y) == Z) && (b(X, W) == Z), (Y == W));
        solver.add(axiom_b_left_cancellative);

        // Axiom 10: ∀X,Y,Z,W. (c(X,Y) = Z ∧ c(W,Y) = Z) ⇒ X = W
        z3::expr axiom_c_right_cancellative = z3::implies((c(X, Y) == Z) && (c(W, Y) == Z), (X == W));
        solver.add(axiom_c_right_cancellative);

        // Axiom 11: ∀X,Y,Z,W. (d(X,Y) = Z ∧ d(W,Y) = Z) ⇒ X = W
        z3::expr axiom_d_right_cancellative = z3::implies((d(X, Y) == Z) && (d(W, Y) == Z), (X == W));
        solver.add(axiom_d_right_cancellative);

        // Conjecture: ∀X,Y,Z,W. (d(X,Y) = Z ∧ d(X,W) = Z) ⇒ Y = W
        // To prove the conjecture, we add its negation and check for unsatisfiability
        z3::expr conjecture_negation = (d(X, Y) == Z) && (d(X, W) == Z) && (Y != W);
        solver.add(!conjecture_negation);

        // Check satisfiability
        z3::check_result result = solver.check();

        if (result == z3::sat) {
            std::cout << "The conjecture is **not** entailed by the axioms (satisfiable)." << std::endl;
            std::cout << "Counterexample:" << std::endl;
            std::cout << solver.get_model() << std::endl;
        } else if (result == z3::unsat) {
            std::cout << "The conjecture is entailed by the axioms (unsatisfiable)." << std::endl;
        } else {
            std::cout << "Solver returned unknown." << std::endl;
        }

        return 0;
    }
    catch (const z3::exception & ex) {
        std::cerr << "Z3 Exception: " << ex.msg() << std::endl;
        return 1;
    }
}
